@@ -21,14 +21,14 @@
 """
 
 import sys
+from os import path
 from inspect import getargspec
 from types import FunctionType
 from re import compile as re_compile
-from os import path
 from collections import MutableMapping, Mapping
 from datetime import timedelta
 try:
-    from yaml import CLoader as Loader
+    from yaml import Loader as Loader
 except ImportError:
     from yaml import Loader
 
@@ -48,10 +48,13 @@ class Configuration(MutableMapping):
     Implements :class:`collections.MutableMapping` protocol.
     """
 
-    def __init__(self, struct=None, ctx=None, parent=None):
+    def __init__(self, struct=None, ctx=None, pwd=None, parent=None):
+        self._pwd = pwd or "."
+        self._parent = parent
         self.__struct = struct
         self.__ctx = ctx or {}
-        self._parent = parent
+
+        self.__ctx["pwd"] = self._pwd
 
     @property
     def _root(self):
@@ -84,7 +87,7 @@ class Configuration(MutableMapping):
             raise ConfigurationError("unconfigured")
         data = self.__struct[name]
         if isinstance(data, dict):
-            return self.__class__(data, self.__ctx, parent=self)
+            return self.__class__(data, self.__ctx, parent=self, pwd=self._pwd)
         if isinstance(data, basestring):
             return data % self.__ctx
         return data
@@ -116,7 +119,7 @@ class Configuration(MutableMapping):
 
     def merge(self, config):
         """ Merge configuration into this one"""
-        new = self.__class__({}, parent=self._parent)
+        new = self.__class__({}, parent=self._parent, pwd=self._pwd)
         new._merge(self)
         new._merge(config)
         return new
@@ -147,14 +150,27 @@ class Configuration(MutableMapping):
     def __add__(self, config):
         return self.merge(config)
 
-    def configure(self, struct):
+    def configure(self, struct=None, _root=True):
         """ Configure with other configuration object"""
-        if isinstance(struct, self.__class__):
-            struct = struct._Configuration__struct
-        self.__struct = struct
+        if struct is not None:
+            if isinstance(struct, self.__class__):
+                struct = struct._Configuration__struct
+            self.__struct = struct
 
-    def configure_obj_graph(self):
-        return configure_obj_graph(self)
+        def _impl(v):
+            if isinstance(v, (Factory, Obj, Include, Extends)):
+                return v(self)
+            if isinstance(v, Configuration):
+                return v.configure()
+            return v
+
+        if _root:
+            if isinstance(self.__struct, Extends):
+                self.__struct = self.__struct(
+                    Configuration.from_dict({}, ctx=self.__ctx, pwd=self._pwd))
+
+        for k, v in self.iteritems():
+            self[k] = _impl(v)
 
     def __repr__(self):
         return repr(self.__struct)
@@ -162,7 +178,7 @@ class Configuration(MutableMapping):
     __str__ = __repr__
 
     @classmethod
-    def from_file(cls, filename, ctx=None, constructors=None):
+    def from_file(cls, filename, ctx=None, pwd=None, constructors=None):
         """ Construct :class:`.Configuration` object by reading and parsing file
         ``filename``.
 
@@ -174,16 +190,14 @@ class Configuration(MutableMapping):
             mapping of names to constructor for custom objects in YAML. Look at
             `_timedelta_constructor` and `_re_constructor` for examples.
         """
+        if pwd is None:
+            pwd = path.dirname(filename)
         with open(filename, "r") as f:
-            cfg = cls(load(f.read(), constructors), ctx=ctx)
-        if "extends" in cfg:
-            supcfg_path = path.join(path.dirname(filename), cfg.pop("extends"))
-            supcfg = cls.from_file(supcfg_path)
-            cfg = supcfg + cfg
-        return cfg
+            return cls.from_string(f.read(), ctx=ctx, pwd=pwd,
+                    constructors=constructors)
 
     @classmethod
-    def from_string(cls, string, ctx=None, constructors=None):
+    def from_string(cls, string, ctx=None, pwd=None, constructors=None):
         """ Construct :class:`.Configuration` from ``string``.
 
         :param string:
@@ -194,10 +208,11 @@ class Configuration(MutableMapping):
             mapping of names to constructor for custom objects in YAML. Look at
             `_timedelta_constructor` and `_re_constructor` for examples.
         """
-        return cls(load(string, constructors), ctx=ctx)
+        cfg = load(string, constructors=constructors)
+        return cls.from_dict(cfg, ctx=ctx, pwd=pwd)
 
     @classmethod
-    def from_dict(cls, d, ctx=None):
+    def from_dict(cls, cfg, ctx=None, pwd=None):
         """ Construct :class:`.Configuration` from dict ``d``.
 
         :param d:
@@ -205,7 +220,7 @@ class Configuration(MutableMapping):
         :param ctx:
             mapping object used for value interpolation
         """
-        return cls(d, ctx=ctx)
+        return cls(cfg, ctx=ctx, pwd=pwd)
 
 def format_config(config, _lvl=0):
     indent = "  " * _lvl
@@ -296,56 +311,6 @@ def configure_logging(logcfg=None, disable_existing_loggers=True):
     from logging.config import dictConfig
     dictConfig(logcfg)
 
-def configure_obj(config, factory=None, ctx=None):
-    config = dict(config)
-    factory = config.pop("factory", None) or factory
-    if not factory:
-        raise ConfigurationError("cannot configure object without factory")
-    if isinstance(factory, basestring):
-        try:
-            factory = import_string(factory)
-        except ImportStringError as e:
-            raise ConfigurationError("cannot import factory: %s" % e)
-    if isinstance(factory, FunctionType):
-        argspec = getargspec(factory)
-    elif isinstance(factory, type):
-        argspec = getargspec(factory.__init__)
-        argspec = argspec._replace(args=argspec.args[1:])
-
-    args = []
-    kwargs = {}
-
-    pos_cut = len(argspec.args) - len(argspec.defaults or [])
-
-    for a in argspec.args[:pos_cut]:
-        if not a in config:
-            raise ConfigurationError(
-                "missing '%s' argument for %s" % (a, factory))
-        arg = config.pop(a)
-        if isinstance(arg, (Ref, Obj)):
-            arg = arg(ctx)
-        args.append(arg)
-
-    for a in argspec.args[pos_cut:]:
-        if a in config:
-            arg = config.pop(a)
-            if isinstance(arg, (Ref, Obj)):
-                arg = arg(ctx)
-            kwargs[a] = arg
-
-    if config:
-        raise ConfigurationError(
-            "extra arguments '%s' found for %s" % (config, factory))
-    return factory(*args, **kwargs)
-
-def configure_obj_graph(config):
-    for k, v in config.iteritems():
-        if isinstance(v, Obj):
-            config[k] = v(config)
-        if isinstance(v, Mapping):
-            config[k] = configure_obj_graph(v)
-    return config
-
 def _timedelta_contructor(loader, node):
     item = loader.construct_scalar(node)
 
@@ -383,25 +348,102 @@ class Ref(object):
 
     def __call__(self, ctx):
         o = ctx.by_ref(self.ref)
-        if isinstance(o, Obj):
+        if isinstance(o, Factory):
             return ctx.by_ref(self.ref, o(ctx))
         return o
 
-def _ref_constructor(loader, node):
-    item = loader.construct_scalar(node)
-    return Ref(item)
+def _ref_constructor(loader, tag, node):
+    return Ref(tag)
 
-class Obj(object):
+class Factory(object):
 
-    def __init__(self, config):
+    def __init__(self, factory, config):
+        self.factory = factory
         self.config = config
 
     def __call__(self, ctx):
-        return configure_obj(self.config, ctx=ctx)
+        config = dict(self.config)
+        factory = self.factory
+        if isinstance(factory, basestring):
+            try:
+                factory = import_string(factory)
+            except ImportStringError as e:
+                raise ConfigurationError("cannot import factory: %s" % e)
+        if isinstance(factory, FunctionType):
+            argspec = getargspec(factory)
+        elif isinstance(factory, type):
+            argspec = getargspec(factory.__init__)
+            argspec = argspec._replace(args=argspec.args[1:])
 
-def _obj_constructor(loader, node):
+        args = []
+        kwargs = {}
+
+        pos_cut = len(argspec.args) - len(argspec.defaults or [])
+
+        for a in argspec.args[:pos_cut]:
+            if not a in config:
+                raise ConfigurationError(
+                    "missing '%s' argument for %s" % (a, factory))
+            arg = config.pop(a)
+            if isinstance(arg, (Ref, Factory)):
+                arg = arg(ctx)
+            args.append(arg)
+
+        for a in argspec.args[pos_cut:]:
+            if a in config:
+                arg = config.pop(a)
+                if isinstance(arg, (Ref, Factory)):
+                    arg = arg(ctx)
+                kwargs[a] = arg
+
+        if config:
+            raise ConfigurationError(
+                "extra arguments '%s' found for %s" % (config, factory))
+        return factory(*args, **kwargs)
+
+def _factory_constructor(loader, tag, node):
     item = loader.construct_mapping(node)
-    return Obj(item)
+    return Factory(tag, item)
+
+class Obj(object):
+
+    def __init__(self, obj):
+        self.obj = obj
+
+    def __call__(self, ctx):
+        try:
+            return import_string(self.obj)
+        except ImportStringError as e:
+            raise ConfigurationError("cannot import obj: %s" % e)
+
+def _obj_constructor(loader, tag, node):
+    return Obj(tag)
+
+class Include(object):
+
+    def __init__(self, filename):
+        self.filename = filename
+
+    def __call__(self, ctx):
+        return Configuration.from_file(path.join(ctx._pwd, self.filename))
+
+def _include_constructor(loader, tag, node):
+    return Include(tag)
+
+class Extends(object):
+
+    def __init__(self, filename, config):
+        self.filename = filename
+        self.config = config
+
+    def __call__(self, ctx):
+        sup = Configuration.from_file(path.join(ctx._pwd, self.filename))
+        cfg = Configuration.from_dict(self.config)
+        return sup + cfg
+
+def _extends_constructor(loader, tag, node):
+    item = loader.construct_mapping(node)
+    return Extends(tag, item)
 
 def load(stream, constructors=None):
     loader = Loader(stream)
@@ -411,10 +453,12 @@ def load(stream, constructors=None):
         loader.add_constructor("!timedelta", _timedelta_contructor)
     if not "re" in constructors:
         loader.add_constructor("!re", _re_constructor)
-    if not "ref" in constructors:
-        loader.add_constructor("!ref", _ref_constructor)
-    if not "obj" in constructors:
-        loader.add_constructor("!obj", _obj_constructor)
+
+    loader.add_multi_constructor("!ref:", _ref_constructor)
+    loader.add_multi_constructor("!factory:", _factory_constructor)
+    loader.add_multi_constructor("!obj:", _obj_constructor)
+    loader.add_multi_constructor("!extends:", _extends_constructor)
+    loader.add_multi_constructor("!include:", _extends_constructor)
 
     if constructors:
         for name, constructor in constructors.items():
